import 'dart:math';

import '../entities/cell.dart';
import '../utils/random_pick_grid.dart';
import '../provider/data.dart';

class BoardUtils {

  static printGrid(List cells) {
    String stringValues = '0123456789';
    print('');
    print('-------');
    for (var rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String row = '';
      for (var colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row += stringValues[cells[rowIndex][colIndex].value];
      }
      print(row);
    }
    print('-------');
    print('');
  }


  static Future<void> pickGrid(Data myProvider) async {
    String grid;
    RandomPickGrid randomPickGrid;

    randomPickGrid = RandomPickGrid();
    await randomPickGrid.init(myProvider.level, myProvider.size);

    if (randomPickGrid.grid != null) {
      grid = randomPickGrid.grid;
    }

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    if (grid.length == pow(blockSizeHorizontal * blockSizeVertical, 2)) {
      print('Picked grid from template: ' + grid);
      bool isSymetric = (blockSizeHorizontal == blockSizeVertical);
      myProvider.updateCells = BoardUtils.createBoardFromTemplate(grid, isSymetric);
      myProvider.selectCell(null, null);
    }
  }


  static List createEmptyBoard(int boardSize) {
    int index = 0;
    List cells = [];
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        row.add(Cell(0, false));
      }
      cells.add(row);
    }

    return cells;
  }


  static List createBoardFromTemplate(String grid, bool isSymetric) {
    List cells = [];
    int boardSize = int.parse(pow(grid.length, 1/2).toStringAsFixed(0));

    String stringValues = '0123456789';

    int index = 0;
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        String stringValue = grid[index++];
        int value = stringValues.indexOf(stringValue);
        row.add(Cell(value, (value != 0)));
      }
      cells.add(row);
    }

    List<String> allowedFlip = ['none', 'horizontal', 'vertical'];
    List<String> allowedRotate = ['none', 'left', 'right'];

    // Forbid rotation if blocks are not symetric
    if (!isSymetric) {
      allowedRotate = ['none'];
    }

    var rand = new Random();
    String flip = allowedFlip[rand.nextInt(allowedFlip.length)];
    String rotate = allowedRotate[rand.nextInt(allowedRotate.length)];

    print('flip board: ' + flip);
    print('rotate board: ' + rotate);

    switch(flip) {
      case 'horizontal': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[boardSize - rowIndex - 1][colIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'vertical': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[rowIndex][boardSize - colIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    switch(rotate) {
      case 'left': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[colIndex][boardSize - rowIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'right': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[boardSize - colIndex - 1][rowIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    // Fix cells fixed states
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        cells[rowIndex][colIndex].isFixed = (cells[rowIndex][colIndex].value != 0) ? true : false;
      }
    }

    printGrid(cells);

    return cells;
  }


  static bool checkBoardIsSolved(Data myProvider) {
    List cells = myProvider.cells;

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    int boardSize = blockSizeHorizontal * blockSizeVertical;

    bool isSolved = true;

    // (re)compute conflicts
    BoardUtils.computeConflictsInBoard(myProvider);

    // check grid is fully completed and does not contain conflict
    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        if (cells[row][col].value == 0 || cells[row][col].conflictsCount != 0) {
          return false;
        }
      }
    }

    print('-> ok sudoku solved!');

    return true;
  }


  static void computeConflictsInBoard(Data myProvider) {
    List cells = myProvider.cells;

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    int boardSize = blockSizeHorizontal * blockSizeVertical;

    // reset conflict states
    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        cells[row][col].conflictsCount = 0;
      }
    }

    // check lines does not contains a value twice
    for (var row = 0; row < boardSize; row++) {
      List values = [];
      for (var col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('line ' + row.toString() + ' contains duplicates');
        // Add line to cells in conflict
        for (var col = 0; col < boardSize; col++) {
          cells[row][col].conflictsCount++;
        }
      }
    }

    // check columns does not contains a value twice
    for (var col = 0; col < boardSize; col++) {
      List values = [];
      for (var row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('column ' + col.toString() + ' contains duplicates');
        // Add column to cells in conflict
        for (var row = 0; row < boardSize; row++) {
          cells[row][col].conflictsCount++;
        }
      }
    }

    // check blocks does not contains a value twice
    int horizontalBlocksCount = blockSizeVertical;
    int verticalBlocksCount = blockSizeHorizontal;
    for (var blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (var blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        List values = [];

        for (var rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (var colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            int row = (blockRow * blockSizeVertical) + rowInBlock;
            int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          print('block [' + blockCol.toString() + ',' + blockRow.toString() + '] contains duplicates');
          // Add blocks to cells in conflict
          for (var rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
            for (var colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
              int row = (blockRow * blockSizeVertical) + rowInBlock;
              int col = (blockCol * blockSizeHorizontal) + colInBlock;
              cells[row][col].conflictsCount++;
            }
          }
        }
      }
    }
  }

}
